@@ -27,6 +27,13 @@ func ErrorAtLeastOneWrap() error {
 	return fmt.Errorf("%v, %w, %v", err1, err2, err3)
 }
 
+func ErrorAtLeastOneWrapWithCustomError() error {
+	err1 := errors.New("oops1")
+	err2 := MyError{}
+	err3 := errors.New("oops3")
+	return fmt.Errorf("%v, %w, %v", err1, err2, err3)
+}
+
 func ErrorStringFormat() error {
 	err := errors.New("oops")
 	return fmt.Errorf("error: %s", err.Error()) // want "non-wrapping format verb for fmt.Errorf. Use `%w` to format errors"
@@ -47,3 +54,13 @@ type MyError struct{}
 func (MyError) Error() string {
 	return "oops"
 }
+
+func ErrorIndexReset() error {
+	err := errors.New("oops1")
+	return fmt.Errorf("%[1]v %d %f %[1]v, %d, %f", err, 1, 2.2) // want "non-wrapping format verb for fmt.Errorf. Use `%w` to format errors"
+}
+
+func ErrorIndexResetGood() error {
+	err := errors.New("oops1")
+	return fmt.Errorf("%[1]w %d %f %[1]v, %d, %f", err, 1, 2.2)
+}
